#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifndef LFATAL
#include "lib_log.h"
#endif

#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/types.h>
#include <fuse.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/un.h>


#include <tcutil.h>
#include <tcbdb.h>
#include <tchdb.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <aio.h>
#include <mhash.h>
#include <mutils/mhash.h>

#include "lib_safe.h"
#include "lib_cfg.h"
#include "lib_net.h"
#include "retcodes.h"
#ifdef LZO
#include "lib_lzo.h"
#endif
#include "lib_qlz.h"
#include "lib_tc.h"
#include "lib_repl.h"
#include "lib_crypto.h"
#include "file_io.h"

extern TCHDB *dbb;
extern TCHDB *dbu;
extern TCHDB *dbp;
extern TCBDB *dbl;
extern TCHDB *dbs;
extern TCHDB *dbdta;
extern TCBDB *dbdirent;
extern TCBDB *freelist;
extern TCTREE *workqtree;
extern TCTREE *readcachetree;
extern int fdbdta;
extern int BLKSIZE;
extern unsigned long long nextoffset;

unsigned char *crc32(unsigned char *buf, int size, int thread_number)
{
    MHASH td[MAX_ALLOWED_THREADS];
    unsigned char *hash[thread_number];

    td[thread_number] = mhash_init(MHASH_CRC32);
    if (td[thread_number] == MHASH_FAILED) exit(1);

    mhash(td[thread_number], buf, size);
    hash[thread_number] = mhash_end(td[thread_number]);
    return hash[thread_number];
}


// In the future we will spawn a session for every thread so we can get rid of the mutex lock
// For now, keep it simple..
void send_repl_data(unsigned char db, unsigned char op, char *key, int ksize, char *value, int vsize, int threadnr)
{
  char *replicationmsg;
  char *msgstart;
  int replicationmsg_size;
  int result;
  int ret=NAK;
  char *crc; 
  int retry=0;

  FUNC;
  repl_lock((char *)__PRETTY_FUNCTION__);
// replication_msg : db (char), op (char), ksize (int),vsize (int), key (char *),value (char *), crc32 (int)
  replicationmsg_size=ksize+vsize+(sizeof(unsigned char) * 2)+(sizeof(int)*2)+sizeof(int);
  replicationmsg=s_zmalloc(replicationmsg_size);
  replicationmsg[0]=db;
  replicationmsg[1]=op;
  memcpy(replicationmsg+(sizeof(unsigned char) * 2),&ksize,sizeof(int));
  memcpy(replicationmsg+(sizeof(unsigned char) * 2)+sizeof(int),&vsize,sizeof(int));
  memcpy(replicationmsg+(sizeof(unsigned char) * 2)+(sizeof(int)*2),key,ksize);
  if ( op == REPLWRITE || op == REPLDUPWRITE || op == REPLDELETECURKEY ) {
    memcpy(replicationmsg+ksize+(sizeof(unsigned char) * 2)+(2 * sizeof(int)),value,vsize);
// For safety we will calc a crc32 checksum of the message.
// drbd has seen cases where data received was garbled despite the
// use of tcp as transfer protocol.
  }

  crc=crc32(replicationmsg,replicationmsg_size-sizeof(int), threadnr);
  memcpy(replicationmsg+replicationmsg_size-sizeof(int),crc,sizeof(int));
  //  memcpy(replicationmsg+ksize+(sizeof(unsigned char) * 2)+vsize+(sizeof(int)*2),crc,sizeof(int));
  //} else {
  //  crc=crc32(replicationmsg,replicationmsg_size-sizeof(int), threadnr);
  //  memcpy(replicationmsg+replicationmsg_size-sizeof(int),crc,sizeof(int));
  //}
  free(crc);

  //repl_lock((char *)__PRETTY_FUNCTION__);
// First send message start and size.
  msgstart=as_sprintf("START : %i",replicationmsg_size); 
start_again:
  retry++;
  result=timeoutWrite(3, config->replication_socket, msgstart, strlen(msgstart)+1);
  if ( result != strlen(msgstart)+1 ) {
     reconnect();
     goto start_again;
  } 
  result=timeoutRead(3, config->replication_socket, (char *)&ret, sizeof(unsigned int));
  if ( ret != replicationmsg_size ) goto start_again;

  result=timeoutWrite(3, config->replication_socket, replicationmsg, replicationmsg_size);
  if ( result != replicationmsg_size ) {
     LFATAL("Failed to write the message, reset connection");
     reconnect();
     goto start_again;
  }
  result=timeoutRead(3, config->replication_socket, (char *)&ret, sizeof(unsigned int));
  if ( ret != ACK ) {
      LINFO("ret != ACK database=%i, operation=%i",db,op);
      goto start_again;
  }
  free(replicationmsg);
  free(msgstart);
  release_repl_lock();
  if ( retry > 1 ) LINFO("ret == ACK database=%i, operation=%i, after %i retries",db,op, retry);
  EFUNC;
  return;
}

void reconnect()
{
    int delay=0;
    close(config->replication_socket);
    while (1) {
        config->replication_socket=clientconnect(config->replication_partner_ip, config->replication_partner_port, "tcp");
        if ( config->replication_socket != -1 ) break;
        sleep(3);
        LFATAL("Retry connect to %s:%s",config->replication_partner_ip, config->replication_partner_port);
    }
    return;
}


void *replication_worker(void *arg)
{
    int msocket;
    int csocket;
    const char *proto = "tcp";
    struct sockaddr_un client_address;
    socklen_t client_len;
    char *message = NULL;
    int result;
    int msglen=0;
    int confirm;
    char *p;
    int crc;
    char *crcmsg;
    int crcm;

    msocket = -1;
    if (NULL == config->replication_listen_port )
       config->replication_listen_port = "101";
    if (NULL == config->replication_listen_ip ) {
       LFATAL("LISTEN_REPLICATION_IP is not specified");
       pthread_exit(NULL);
    }

    while (1) {
        msocket = serverinit(config->replication_listen_ip, config->replication_listen_port, proto);
        if (msocket != -1)
            break;
        sleep(1);
        close(msocket);
    }
    message=s_zmalloc(BLKSIZE *2); // Enough space to hold the message and the kitchen sink
    client_len = sizeof(client_address);
    while (1) {
        config->replication_socket =
            accept(msocket, (struct sockaddr *) &client_address,
                   &client_len);
        while (1) {
           result=readstring(1, config->replication_socket, message, 64);
           if ( result == 0 ) {
                if ( 0 != strncmp("START :",message,strlen("START :"))) continue;
                p=strchr(message,':');
                *p++;
                msglen=atoi(p);
           } else continue;
           result=timeoutWrite(3, config->replication_socket, (char *)&msglen,sizeof(int));
           if ( result != sizeof(int)) {
              LINFO("Slave failed to send ACK, disconnect");
              close(config->replication_socket);
              break;
           }
           result=fulltimRead(10,config->replication_socket,message,msglen);
           if ( result != msglen) {
              LINFO("Slave : got %i expected %i",result,msglen);
              send_nak();
              continue;
           }
           crcmsg=crc32(message,msglen-sizeof(int), 1);    
           memcpy(&crcm,crcmsg,sizeof(int));
           free(crcmsg);
           memcpy(&crc,&message[msglen-sizeof(int)],sizeof(int));
           if ( crc != crcm ) {
              LINFO("crc (%i) != crcmsg (%i)",crc,crcm);
              send_nak();
              continue;
           }
           send_ack();
           process_message(message, msglen-sizeof(int));
        }
    }
// We never get here.
    free(message);
    pthread_exit(NULL);
}

void process_message(char *message, int msglen)
{
//message : db (char), op (char), ksize (int),vsize (int), key (char *),value (char *)
  REPLICATIONMSG replicationmsg;
  unsigned long long offset;

  replicationmsg.database=message[0];
  replicationmsg.operation=message[1];
  memcpy(&replicationmsg.ksize,&message[2],sizeof(int));
  memcpy(&replicationmsg.vsize,&message[2+sizeof(int)],sizeof(int));
  replicationmsg.key=&message[2+(2*sizeof(int))];
  replicationmsg.value=&message[2+(2*sizeof(int))+replicationmsg.ksize];

  if ( replicationmsg.operation == REPLWRITE ) {
     LDEBUG("process_message : write keylen %i, vsize %i",replicationmsg.ksize, replicationmsg.vsize);
     if ( replicationmsg.database == DBDTA ) {
        bin_write_dbdata(dbdta,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize); 
     }
     if ( replicationmsg.database == DBU ) {
        bin_write_dbdata(dbu,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize); 
     }
     if ( replicationmsg.database == DBB ) {
        bin_write_dbdata(dbb,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize); 
     }
     if ( replicationmsg.database == DBP ) {
        bin_write_dbdata(dbp,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize); 
     }
     if ( replicationmsg.database == DBS ) {
        bin_write_dbdata(dbs,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize); 
     }
     if ( replicationmsg.database == DBDIRENT ) {
        btbin_write_dbdata(dbdirent,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize); 
     }
     if ( replicationmsg.database == FREELIST ) {
        btbin_write_dbdata(freelist,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize); 
     }
     if ( replicationmsg.database == DBL ) {
        btbin_write_dbdata(dbl,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize); 
     }
     if ( replicationmsg.database == FDBDTA ) {
// With a write to fdbdta we use the key as store for the offset.
        memcpy(&offset,replicationmsg.key,replicationmsg.ksize);
        s_lckpwrite(fdbdta, replicationmsg.value, replicationmsg.vsize, offset);
     }
  }
  if ( replicationmsg.operation == REPLDUPWRITE ) {
    if ( replicationmsg.database == DBDIRENT ) {
        btbin_write_dup(dbdirent,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize);
     }
     if ( replicationmsg.database == FREELIST ) {
        btbin_write_dup(freelist,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize);
     }
     if ( replicationmsg.database == DBL ) {
        btbin_write_dup(dbl,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize);
     }
  }
  if ( replicationmsg.operation == REPLDELETE ) {
     if ( replicationmsg.database == DBDTA ) {
        delete_key(dbdta,replicationmsg.key,replicationmsg.ksize);
     }
     if ( replicationmsg.database == DBU ) {
        delete_key(dbu,replicationmsg.key,replicationmsg.ksize);
     }
     if ( replicationmsg.database == DBB ) {
        delete_key(dbb,replicationmsg.key,replicationmsg.ksize);
     }
     if ( replicationmsg.database == DBP ) {
        delete_key(dbp,replicationmsg.key,replicationmsg.ksize);
     }
     if ( replicationmsg.database == DBS ) {
        delete_key(dbs,replicationmsg.key,replicationmsg.ksize);
     }
  }
  if ( replicationmsg.operation == REPLDELETECURKEY ) {
     if ( replicationmsg.database == DBDIRENT ) {
        btdelete_curkey(dbdirent,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize);
     }
     if ( replicationmsg.database == FREELIST ) {
        btdelete_curkey(freelist,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize);
     }
     if ( replicationmsg.database == DBL ) {
        btdelete_curkey(dbl,replicationmsg.key,replicationmsg.ksize,replicationmsg.value,replicationmsg.vsize);
     }
  }
  if ( replicationmsg.operation == REPLSETNEXTOFFSET ) {
     memcpy(&nextoffset,replicationmsg.key,replicationmsg.ksize);
  }
  return;
}

int send_nak()
{
  int confirm=NAK;
  int result;
  LFATAL("send_nak");
  result=timeoutWrite(3, config->replication_socket, (char *)&confirm,sizeof(int));
  if ( result != sizeof(unsigned int)) {
     close(config->replication_socket);
     return(-1);
  }
  return(0);
}

int send_ack()
{
  int confirm=ACK;
  int result;
  //LFATAL("send_ack");
  result=timeoutWrite(3, config->replication_socket, (char *)&confirm,sizeof(int));
  if ( result != sizeof(unsigned int)) {
     close(config->replication_socket);
     return(-1);
  }
  return(0);
}

